#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
#_plots.py

Module (graph): .Mixin functions to inherit, plotting functions that use instance attributes from Calcium Data
 class and nothing else.
"""
from __future__ import annotations

import logging
from typing import Optional, Iterable, Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from canalysis.graphs.graph_utils import ax_helpers

logger = logging.getLogger(__name__)

class CalPlots:

    tracedata: Any
    doevents: Any
    doeating: Any
    eventdata: Optional[Any]
    eatingdata: Optional[Any]
    signals: pd.DataFrame | pd.Series | np.ndarray | Iterable
    time: np.ndarray | Iterable | int | float | bool
    cells: np.ndarray | pd.Series | Iterable | int | float | bool
    session: str
    trial_times: dict
    timestamps: dict
    color_dict: dict

    def plot_stim(self, save_dir: str = None) -> None:
        if save_dir:
            save_dir = save_dir
        nplot = len(self.signals.columns)
        for stim, times in self.trial_times.items():
            trialno = 0
            for trial in times:
                trialno += 1
                # get only the data within the analysis window
                data_ind = np.where((self.time > trial - 2) & (self.time < trial + 5))[
                    0
                ]
                # index to analysis data
                this_time = self.time[data_ind]

                fig, ax = plt.subplots(nplot, 1, sharex=True)
                for i in range(0, nplot):
                    # Get calcium trace for this analysis window
                    signal = list(self.signals.iloc[data_ind, i])
                    # plot signal
                    ax[i].plot(this_time, signal, "k", linewidth=1)
                    ax[i].get_xaxis().set_visible(False)
                    ax[i].spines["top"].set_visible(False)
                    ax[i].spines["bottom"].set_visible(False)
                    ax[i].spines["right"].set_visible(False)
                    ax[i].set_yticks([])
                    ax[i].set_ylabel(
                        self.signals.columns[i],
                        rotation="horizontal",
                        labelpad=15,
                        y=0.1,
                    )
                    # Add shading.
                    for stimmy in ["Lick", "Rinse", stim]:
                        done = 0
                        timey = self.timestamps[stimmy]
                        for ts in timey:
                            if trial - 1 < ts < trial + 3:
                                if done == 0:
                                    label = stimmy
                                    done = 1
                                else:
                                    label = "_"
                                ax[i].axvspan(
                                    ts,
                                    ts + 0.15,
                                    color=self.color_dict[stimmy],
                                    label=label,
                                    lw=0,
                                )

                # Make the plots act like they know each other.
                fig.subplots_adjust(hspace=0)
                plt.xlabel("Time (s)")
                ax[-1].get_xaxis().set_visible(True)
                ax[-1].spines["bottom"].set_visible(True)
                fig.set_figwidth(4)
                if save_dir:
                    fig.savefig(
                        save_dir + f"/{stim}_{trial}.png",
                        bbox_inches="tight",
                        dpi=600,
                        facecolor="white",
                    )
        return None

    def plot_session(self, lickshade: int = 1, save: bool = False, eatingdata=None) -> None:
        # create a series of plots with a shared x-axis
        fig, ax = plt.subplots(
            len(self.tracedata.cells), 1, sharex=True, facecolor="white"
        )
        for i in range(len(self.tracedata.cells)):
            # get calcium trace (y axis data)
            signal = list(self.tracedata.signals.iloc[:, i])
            # plot signal
            ax[i].plot(self.tracedata.time, signal, "k", linewidth=0.8)
            # Get rid of borders
            ax[i].get_xaxis().set_visible(False)
            ax[i].spines["top"].set_visible(False)
            ax[i].spines["bottom"].set_visible(False)
            ax[i].spines["right"].set_visible(False)
            ax[i].set_yticks([])  # no Y ticks
            # add the cell name as a label for this graph's y-axis
            ax[i].set_ylabel(
                self.tracedata.signals.columns[i],
                rotation="horizontal",
                labelpad=15,
                y=0.1,
                fontweight="bold",
            )
            # shade in licks
            if self.doevents:
                for lick in self.eventdata.timestamps["Lick"]:
                    label = "_yarp"
                    if lick == self.eventdata.timestamps["Lick"][0]:
                        label = "Licking"
                    ax[i].axvspan(
                        lick, lick + lickshade, color="lightgray", lw=0, label=label
                    )
            if self.doeating:
                for interv in self.eatingdata.raw_eatingdata.reset_index(drop=True).to_numpy():
                    if interv[0] == 'Grooming':
                        ax[i].axvspan(
                                interv[1], interv[2], color='cyan', lw=0, label=interv[0]
                        )
                        logging.info(f"placing grooming interv {interv[1]}-{interv[0]}")
                    if interv[0] == 'EATING':
                        ax[i].axvspan(
                                interv[1], interv[2], color='blue', lw=0, label=interv[0]
                        )
            ax[i].yaxis.label.set_fontsize(10)

        fig.subplots_adjust(hspace=0)
        ax[-1].get_xaxis().set_visible(True)
        ax[-1].spines["bottom"].set_visible(True)
        # ax_helpers.make_legend(
        #         {
        #             "Eating" : "blue",
        #             "Grooming" : "cyan",
        #             "Licking" : "lightgray"
        #         },
        #         marker="s",
        # )
        plt.show()
        if save:
            fig.savefig(
                f"C:\\Users\\flynn\\Desktop\\figs\\{self.session}_session.png",
                bbox_inches="tight",
                dpi=1000,
                facecolor="white",
            )
        return None

    def plot_zoom(
        self,
        zoomshade: float = 0.2,
        save: Optional[bool] = True,
    ) -> None:
        # create a series of plots with a shared x-axis
        fig, ax = plt.subplots(len(self.tracedata.cells), 1, sharex=True)
        # zoombounding = [
        #     int(input("Enter start time for zoomed in graph (seconds):")),
        #     int(input("Enter end time for zoomed in graph (seconds):")),
        # ]
        zoombounding = [
            150,
            245,
        ]
        for i in range(len(self.tracedata.cells)):
            signal = list(self.tracedata.signals.iloc[:, i])
            ax[i].plot(self.tracedata.time, signal, "k", linewidth=0.8)
            ax[i].get_xaxis().set_visible(False)
            ax[i].spines["top"].set_visible(False)
            ax[i].spines["bottom"].set_visible(False)
            ax[i].spines["right"].set_visible(False)
            ax[i].set_yticks([])
            ax[i].set_ylabel(
                self.tracedata.signals.columns[i], rotation="horizontal", labelpad=15, y=0.1
            )
            # shade in timestamps
            if self.doevents:
                for stim, times in self.eventdata.timestamps.items():
                    if stim != "ArtSal":
                        for ts in times:
                            if ts == times[0]:
                                label = stim
                            else:
                                label = "_"  # Keeps label from showing.
                            ax[i].axvspan(
                                ts,
                                ts + zoomshade,
                                color=self.color_dict[stim],
                                label=label,
                                lw=0,
                            )
            if self.doeating:
                for interv in self.eatingdata.raw_eatingdata.to_numpy():
                    if interv[0] == 'Grooming':
                        ax[i].axvspan(
                                interv[1], interv[2], color='cyan', lw=0, label=interv[0]
                        )
                    if interv[0] == 'EATING':
                        ax[i].axvspan(
                                interv[1], interv[2], color='blue',alpha=0.2, lw=0, label=interv[0]
                        )
                    if interv[0] == 'BackLeft':
                        ax[i].axvspan(
                                interv[1]-0.8, interv[2], color='lime', lw=0, label=interv[0]
                        )
            ax[i].yaxis.label.set_fontsize(10)
        fig.subplots_adjust(hspace=0)
        plt.xlabel("Time (s)")
        ax[-1].get_xaxis().set_visible(True)
        ax[-1].spines["bottom"].set_visible(True)
        ax_helpers.make_legend(
                {
                    "Apple - Acquisition": "lime",
                    "Apple - Eating": "blue",
                },
                marker="s",
                markeralpha=0.2
        )
        # set the x-axis to the zoomed area
        plt.setp(ax, xlim=zoombounding)
        plt.show()
        if save:
            fig.savefig(
                f"C:\\Users\\flynn\\Desktop\\figs\\{self.session}_zm.png",
                bbox_inches="tight",
                dpi=1200,
            )
        return None

    def plot_cells(self, save_dir: Optional[bool] = True) -> None:
        cells = self.cells
        if cells is None:
            cells = self.cells

        for cell in cells:
            # Create int used to index a column in tracedata.
            cell_index = {x: 0 for x in cells}
            cell_index.update((key, value) for value, key in enumerate(cell_index))
            currcell = cell_index[cell]
            for stim, times in self.trial_times.items():
                ntrial = len(times)
                minmax = []
                # Max/mins to standardize plots
                for it, tri in enumerate(times):
                    temp_data_ind = np.where(
                        (self.time > tri - 2) & (self.time < tri + 5)
                    )[0]
                    temp_signal = self.tracedata.iloc[temp_data_ind, currcell + 1]
                    norm_min = min(temp_signal)
                    norm_max = max(temp_signal)
                    minmax.append(norm_min)
                    minmax.append(norm_max)
                stim_min = min(minmax)
                stim_max = max(minmax)
                fig, xaxs = plt.subplots(ntrial, 1, sharex=False, squeeze=False)
                if ntrial < 2:
                    xaxs.flatten()
                for iteration, trial in enumerate(times):
                    i = int(iteration)
                    data_ind = np.where(
                        (self.time > trial - 2) & (self.time < trial + 4)
                    )[0]
                    this_time = self.time[data_ind]
                    signal = list(self.tracedata.iloc[data_ind, currcell + 1])
                    signal[:] = [number - stim_min for number in signal]
                    l_bound = min(signal)
                    u_bound = max(signal)
                    center = 0
                    xaxs[i, 0].plot(this_time, signal, "k", linewidth=0.8)
                    xaxs[i, 0].tick_params(axis="both", which="minor", labelsize=6)
                    xaxs[i, 0].get_xaxis().set_visible(False)
                    xaxs[i, 0].spines["top"].set_visible(False)
                    xaxs[i, 0].spines["bottom"].set_visible(False)
                    xaxs[i, 0].spines["right"].set_visible(False)
                    xaxs[i, 0].spines["left"].set_bounds((l_bound, stim_max))
                    xaxs[i, 0].set_yticks((0, center, u_bound))
                    xaxs[i, 0].set_ylabel(
                        " Trial {}     ".format(i + 1),
                        rotation="horizontal",
                        labelpad=15,
                        y=0.3,
                    )
                    xaxs[i, 0].set_ylim(bottom=0, top=max(signal))
                    xaxs[i, 0].axhspan(0, 0, color="k", ls=":")
                    # Add shading for licks, rinses  tastant delivery
                    for stimmy in ["Lick", "Rinse", stim]:
                        done = 0
                        timey = self.timestamps[stimmy]
                        for ts in timey:
                            if trial - 1.5 < ts < trial + 5:
                                if done == 0:
                                    label = stimmy
                                    done = 1
                                else:
                                    label = "_"
                                xaxs[i, 0].axvspan(
                                    ts,
                                    ts + 0.15,
                                    color=self.color_dict[stimmy],
                                    label=label,
                                    lw=0,
                                )
                plt.xlabel("Time (s)")
                fig.suptitle(
                    "Calcium Traces: {}\n{}: {}".format(cell, self.session, stim), y=1.0
                )
                fig.set_figwidth(6)
                fig.text(
                    0,
                    -0.03,
                    "Note: Each trace has" "been normalized over the graph window",
                    fontstyle="italic",
                    fontsize="small",
                )
                xaxs[-1, 0].spines["bottom"].set_bounds(False)
                plt.legend(loc=(1.02, 3))
        if save_dir:
            plt.savefig(str(save_dir) + ".png", bbox_inches="tight", dpi=600)
        return None




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 19 20:39:48 2022

@author: flynnoconnell
"""
from . import _properties
from . import _validate
from . import datahandler
from . import funcs
from . import scores



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
#wrappers.py 

Module (utils): Wrappers to decorate for functionality.
"""

import functools
import time


def log_time(func):
    """Logs the time it took for func to execute"""
    def wrapper(*args, **kwargs):
        start = time.time_ns() / (10 ** 9)
        val = func(*args, **kwargs)
        end = time.time_ns() / (10 ** 9)
        duration = end - start
        print(f"{func.__name__} execution time: {duration}s")
        return val
    return wrapper


def typecheck(*args1, isclassmethod=False):
    """
    This typecheck function decorator to declare the type of each function or method argument. 
    
    Example:
        @typecheck(types.StringType, Decimal)
        def my_function(s, d):
            pass
        
    Types: https://stackless.readthedocs.io/en/2.7-slp/library/types.html
    Cannot be used to typecheck n method argument that is of the same type as the method's
    class.

    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args2, **keywords):
            args = args2[1:] if isclassmethod else args2
            for (arg2, arg1) in zip(args, args1):
                if not isinstance(arg2, arg1):
                    raise TypeError(f"Expected type: {arg1}, actual type: {type(arg2)}")
            return func(*args2, **keywords)

        return wrapper

    return decorator


class Singleton:
    """
    A non-thread-safe helper class to ease implementing singletons.
    This should be used as a decorator -- not a metaclass -- to the
    class that should be a singleton.
    The decorated class can define one `__init__` function that
    takes only the `self` argument. Also, the decorated class cannot be
    inherited from. Other than that, there are no restrictions that apply
    to the decorated class.
    To get the singleton instance, use the `Instance` method. Trying
    to use `__call__` will result in a `TypeError` being raised.
    Ref: https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """

    def __init__(self, decorated):
        self._decorated = decorated

    def Instance(self):
        """
        Returns the singleton instance. Upon its first call, it creates a
        new instance of the decorated class and calls its `__init__` method.
        On all subsequent calls, the already created instance is returned.
        """
        try:
            return self._instance
        except AttributeError:
            self._instance = self._decorated()
            return self._instance

    def __call__(self):
        raise TypeError("Singletons must be accessed through `Instance()`.")

    def __instancecheck__(self, inst):
        return isinstance(inst, self._decorated)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
#taste_data
"""
import logging
from typing import Generator, Iterable, Optional, Tuple
import pandas as pd
import numpy as np
from dataclasses import dataclass

from canalysis.helpers import funcs

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(name)s - %(message)s")


@dataclass
class TasteData:
    __signals: pd.DataFrame
    __time: pd.Series
    __timestamps: dict
    trial_times: dict
    __avgs: dict
    color_dict: dict
    baseline: int = 0
    post: int = 4

    def __post_init__(self):
        assert isinstance(self.__signals, pd.DataFrame)
        self.tastedata = self.concat_event_signals(self.__timestamps).drop(columns=['time'])
        self.events: pd.Series = self.tastedata['event']
        self.colors: pd.Series = self.tastedata['color']
        self.signals = self.tastedata.drop(columns=['event', 'color'])

    def __repr__(self):
        return type(self).__name__

    def concat_event_signals(self, timestamps):
        """
        From data.timestamps, iterate each event and get signals
        starting from interval[0] to interval [1].

        Parameters
        ----------
        self : instance
        timestamps : dict
            Timestamps to sort through.
        Returns
        -------
        method
            Call to method where specific attributes are set.
        """
        logging.info("Setting taste data...")
        aggregate_signals_df = pd.DataFrame()
        for event, interv in funcs.iter_events(timestamps):
            if event not in ['Lick', 'ArtSal']:
                event_df = self.__signals.loc[
                    (self.__time > (interv[0] - self.baseline)) &
                    (self.__time < (interv[1] + self.post))].copy()
                event_df['color'] = self.color_dict[event]
                event_df["event"] = event
                aggregate_signals_df = pd.concat([
                    aggregate_signals_df, event_df],
                    axis=0)
        logging.info("Taste data set.")
        return aggregate_signals_df

    def get_signals_from_events(self, events: list) -> Tuple[pd.DataFrame, pd.Series]:
        signal = self.tastedata[self.tastedata['event'].isin(events)].drop(columns=['event'])
        color = signal.pop('color')
        return signal, color

    def get_taste_df(
            self, zero: bool=True,
    ) -> Generator[Iterable, None, None]:
        signals = self.__signals.drop("time", axis=1)
        if zero:
            for cell in signals:
                signals[cell] = signals[cell] - self.__avgs[cell]
                # Replace negatives with 0 using numpys fancy indexing
                signals[cell][signals[cell] < 0] = 0

        for stim, times in self.trial_times.items():
            for iteration, trial in enumerate(times):
                data_ind = np.where((self.__time > trial-2) & (self.__time < trial + 5))[0]
                signal = signals.iloc[data_ind, :]
                yield stim, iteration, signal

    def loop_taste(
            self,
            save_dir: Optional[str] = "",
            **kwargs
    ) -> Generator[Iterable, None, None]:
        from heatmaps import EatingHeatmap
        for stim, iteration, signal in self.get_taste_df():
            heatmap = EatingHeatmap(
                    signal.T,
                    title=f'{stim}, Trial: {iteration + 1}',
                    save_dir=save_dir,
                    save_name=f'{stim}_{iteration}',
                    **kwargs
            )
            fig = heatmap.default_heatmap(maptype='taste')
            yield fig



"""
# displayable_path.py

Module(utils/file_helpers): Class for displaying path tree.
"""
from __future__ import annotations

from pathlib import Path


class DisplayablePath:
    """
    Class for creating a tree to display current directory.

    from https://stackoverflow.com/a/49912639

    Parameters
    ----------
    path : str
       - path to innermost dir
    parent_path : str | Path
       - path to outermost dir

    Example Usage
    ----------
    paths = DisplayablePath.make_tree(
       Path('doc'),
       criteria=is_not_hidden
    )
    for path in paths:
       print(path.displayable())

    # With a criteria (skip hidden files)
    def is_not_hidden(path):
       return not path.name.startswith(".")

    paths = DisplayablePath.make_tree(Path('doc'))
    for path in paths:
       print(path.displayable())

    Example Output:
    ----------
    doc/
    ├── _static/
    │   ├── embedded/
    │   │   ├── deep_file
    │   │   └── very/
    │   │       └── deep/
    │   │           └── folder/
    │   │               └── very_deep_file
    │   └── less_deep_file
    ├── about.rst
    ├── conf.py
    └── index.rst
    """

    display_filename_prefix_middle = "├──"
    display_filename_prefix_last = "└──"
    display_parent_prefix_middle = "    "
    display_parent_prefix_last = "│   "

    def __init__(self, path: str, parent_path, is_last) -> None:

        self.path: str | Path = Path(str(path))
        self.parent = parent_path
        self.is_last = is_last
        if self.parent:
            self.depth = self.parent.depth + 1
        else:
            self.depth = 0

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + "/"
        return self.path.name

    @classmethod
    def make_tree(cls, root, parent=None, is_last=False, criteria=None):
        root = Path(str(root))
        criteria = criteria or cls._default_criteria

        displayable_root = cls(root, parent, is_last)
        yield displayable_root

        children = sorted(
            list(path for path in root.iterdir() if criteria(path)),
            key=lambda s: str(s).lower(),
        )
        count = 1
        for path in children:
            is_last = count == len(children)
            if path.is_dir():
                yield from cls.make_tree(
                    path, parent=displayable_root, is_last=is_last, criteria=criteria
                )
            else:
                yield cls(path, displayable_root, is_last)
            count += 1

    @classmethod
    def _default_criteria(cls, path):
        return True

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + "/"
        return self.path.name

    def displayable(self):
        if self.parent is None:
            return self.displayname

        _filename_prefix = (
            self.display_filename_prefix_last
            if self.is_last
            else self.display_filename_prefix_middle
        )

        parts = ["{!s} {!s}".format(_filename_prefix, self.displayname)]

        parent = self.parent
        while parent and parent.parent is not None:
            parts.append(
                self.display_parent_prefix_middle
                if parent.is_last
                else self.display_parent_prefix_last
            )
            parent = parent.parent

        return "".join(reversed(parts))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# datahandler.py

Module (nn_utils): Class to handle data for import into SVM classifier.
"""
import numpy as np


class DataHandler:
    def __init__(self, data, target):
        """
        Check and index specific data to feed into SVM. Accepted as input to sklearn.GridSearchCV().
        Features are the data used for regression and margin vectorizations.
        Labels (or targets, synonymous) are what the classifier is being trained on.

        Parameters
        ----------
        data : pd.DataFrame | np.ndarray
            Features. 
        target : pd.Series | np.ndarray
            Labels/targets.
        """
        assert data.shape[0] == target.shape[0]
        self.data = np.array(data)
        self.target = np.array(target)

    def __getitem__(self, idx: int):
        """
        Indexer. Returns both feature and target values.

        Parameters
        ----------
        idx : int | Iterable[Any]
            Indexors of any type matching the index of the given dataset.

        Returns
        -------
        slice
            Indexed features.
        slice
            Indexed targets.
        """
        return self.data[idx], self.target[idx]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# heatmaps.py
"""

from __future__ import annotations

from . import graph_utils
from .base import _base_heatmap
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from canalysis.helpers import funcs
import logging
logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)
graph_utils.helpers.update_rcparams()



def add_slash(path):
    """
    Adds a slash to the path, but only when it does not already have a slash at the end
    Params: a string
    Returns: a string
    """
    if not path.endswith('/'):
        path += '/'
    return path


class EatingHeatmap(_base_heatmap.BaseHeatmap):
    def __init__(
        self,
        data,
        premask: np.ndarray = None,
        cmap: str = "plasma",
        colorbar: bool = False,
        save_dir: str | None = "",
        save_name: str = "",
        title: str | None = "",
        **figargs
    ):
        super().__init__(cmap, colorbar, **figargs)
        self.data = data
        self.data.columns = np.round(np.arange(0, len(self.data.columns) / 10, 0.1), 1)
        self.premask = premask
        self.title = title
        self.save_dir = save_dir
        self.save_name = save_name
        if self.save_dir is not None and self.save_name is None:
            import random
            self.save_name = str(random.random())
            logging.info(f'No name given to save the file. Name designated: {self.save_name}')
        """
        Create heatmaps.

        .. note::
            -Using seaborn(sns).heatmap(), each row of the heatmap needs to correspond to each row of dictionary.
            -If incoming data_dict is already in this format, you can just delete the df = df.T.
            -Using matplotlib.pyplot(plt).imshow(), each column of the heatmap corresponds to columns of the data.
        Parameters
        ----------
        save_dir: str = ''
            Path to save the file.
        save_name: str = ''
            Name to give the saved file.
        cm: str = 'inferno'
            Colormap color. Options:
             1) https://seaborn.pydata.org/generated/seaborn.color_palette.html?highlight=sns%20color_palette
             2) https://matplotlib.org/stable/tutorials/colors/colormaps.html
        _id : str
            Identifier that will be appended to the filenames this batch of graphs.
                -e.g. _id = 'heatmaps' -> .../heatmaps.png
        title: str = ''
            If parameter passed, set this string as the plot title.
        sigma : scalar or sequence of scalars = 0
            Standard deviation for Gaussian kernal. This smooths out the heatmap to give a more
            uniform distribution. Set to 0 for no smoothing.
        square : bool = False
            If True, set the Axes' aspect to “equal” so each cell will be square-shaped.
        colorbar : bool = False
            If true, a colorbar will be plotted alongside each heatmap.
        robust: bool = False
            If true, and cbar = True, the colormap range is computed with robust quantiles instead of the extreme values
        line_loc : Optional[int] = 0
            If an int is set, plot a vertical(or horizontal) line at the x(or y) value indicated by the given number.
        linewidth : Optional[int] = 3
            Width of vertical line if line_loc is set.
        line_color: str = 'white'
            Color of vertical line, if line_loc is set.
        """

    @property
    def ax(self, ):
        return self._ax

    @ax.setter
    def ax(self, value):
        self._ax = value

    def show(self):
        self.fig.show()

    def save(self,) -> None:
        if self.save_dir is None:
            raise AttributeError('Attempted save without save directory arg.')
        save_dir = add_slash(self.save_dir) # make sure we start in new dir
        filename = f'{save_dir}' + self.save_name + '.png'
        savefile = funcs.check_unique_path(filename)
        self.fig.savefig(f"{savefile}", dpi=400, bbox_inches="tight", pad_inches=0.01, )
        return None

    def default_heatmap(self,
                        line1=None,
                        line2=None,
                        line3=None,
                        maptype: str = 'eating',
                        **kwargs
                        ):
        """Plot single heatmap with seaborn library."""
        self.ax.set_title(self.title, fontweight="bold")
        self.data = self.data[self.data.columns].astype(float)
        self.ax = sns.heatmap(self.data, cbar=self.colorbar, cmap=self.cmap, **kwargs)
        if maptype == 'eating':
            self.set_eatingmap_lines(line1, line2, line3)
        if maptype == 'taste':
            self.set_tastemap_lines()
            self.set_taste_axislabel()
        self.ax.tick_params(axis="x", bottom=True, top=False, labelbottom=True, labeltop=False)
        self.ax.xaxis.set_ticklabels(ticklabels=self.ax.get_xticklabels(), rotation=45)
        self.ax.set_yticks(list(i + 0.5 for i in range(0, self.data.shape[0])))
        self.ax.set_yticklabels(list(self.data.index.values))
        if maptype=='eating':
            n = 2
            [l.set_visible(False) for (i, l) in enumerate(self.ax.xaxis.get_ticklabels()) if i % n != 0]
        if self.save_dir:
            self.save()
        self.fig.close()
        return self.fig

    def set_eatingmap_lines(self, eatingstart, entrystart, eatingend):
        line_loc1 = (eatingstart - entrystart) * 10
        line_loc2 = (eatingend - eatingstart) * 10
        self.ax.axvline(
                line_loc1,
                color='w',
                linewidth=3,)
        self.ax.axvline(
                line_loc2,
                color='w',
                linewidth=3,)
        return None

    def set_tastemap_lines(self):
        self.ax.axvline(
                20,
                color='w',
                linewidth=3,)
        return None

    @staticmethod
    def set_taste_axislabel():
        plt.xticks([0, 20, 69], ['-2', '0', '7'])





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 19 20:39:48 2022
(Module) Data container and data utility functions.
@author: flynnoconnell
"""

import logging

logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)

from .containers import *
from .data_utils import *

__all__ = ["data_utils", "containers"]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# _properties.py

Module(nn_utils): Boilerplate property getter/setters for SVM model.
    
"""
from neuralnetwork.nn_utils._validate import _validate
from neuralnetwork.nn_utils.datahandler import DataHandler
from typing import Any


class _props:
    _evaldata: DataHandler
    _cv: Any

    @property
    def evaldata(self):
        return self._evaldata

    @evaldata.setter
    def evaldata(self, handlerargs):
        data, target, stage = handlerargs

    @property
    def cv(self):
        return self._cv

    @cv.setter
    def cv(self, new_cv):
        self._cv = new_cv

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, new_model):
        if new_model:
            _validate._check_fitted(new_model)
            print(f"Model set: {new_model}")
        self._model = new_model



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# ax_helpers.py

Module (graphs.graph_utils): Functions for manipulating axes objects.

"""
from __future__ import annotations

import logging
from typing import Optional, Any
import numpy as np
from canalysis.graphs.graph_utils import helpers

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(name)s - %(message)s")


def get_axis_labels(ax, data):
    ax.set_xlabel(data.columns[0], weight="bold")
    ax.set_ylabel(data.columns[1], weight="bold")
    if data.shape[1] > 2:
        ax.set_zlabel(data.columns[1], weight="bold")
    return ax


def get_legend(ax, color_dict, colors, markersize):
    mydict = {k: v for k, v in color_dict.items() if color_dict[k] in np.unique(colors)}
    proxy, label = helpers.get_handles_from_dict(mydict, markersize)
    ax.legend(
        handles=proxy,
        labels=label,
        prop={"size": 20},
        bbox_to_anchor=(1.05, 1),
        ncol=2,
        numpoints=1,
        facecolor=ax.get_facecolor(),
        edgecolor=None,
        fancybox=True,
        markerscale=True)
    return ax

def make_legend(
    mydict: dict,
    marker: Optional[str] = 'o',
    show: Optional[bool] = True,
    save: Optional[bool] = True,
    markeralpha = None,
) -> None:
    helpers.update_rcparams()
    logging.info(f"..making legend with marker {marker}")
    import pylab
    fig = pylab.figure()
    figlegend = pylab.figure(figsize=(3, 2))
    ax = fig.add_subplot(111)
    proxy, label = gr_func.get_handles_from_dict(
            mydict,
            markersize=5,
            marker=marker
    )
    leg = figlegend.legend(
            handles=proxy,
            labels=label,
            fancybox=False,
    )
    if markeralpha is not None:
        for lh in leg.legendHandles:
            lh.set_alpha(markeralpha)
    if show:
        fig.show()
        figlegend.show()
        logging.info('Legend showing')
    if save:
        mydir = 'C:\\Users\\flynn\\Desktop\\figs\\legend.png'
        figlegend.savefig(f"{mydir}", dpi=1200)
        logging.info(f'Legend saved in {mydir}')


"""
# draw_plots.py

Module (core): Functions for drawing graphs.

"""
from __future__ import annotations

import logging
from typing import Iterable
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import rcParams


def set_pub():
    rcParams.update(
        {
            "font.weight": "bold",
            "axes.labelweight": "bold",
            "axes.facecolor": "w",
            "axes.labelsize": 15,
            "lines.linewidth": 0.25,
        }
    )

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(name)s - %(message)s")


class QuickPlot:
    def __init__(
        self,
        data: pd.DataFrame,
        time: Iterable[any],
        cmap: str = "magma",
        **kwargs,
    ):
        self.cmap = plt.get_cmap(cmap)
        self.data = data
        self.time = time

        self.facecolor = "white"
        self.color_dict = {
            "ArtSal": "dodgerblue",
            "MSG": "darkorange",
            "NaCl": "lime",
            "Sucrose": "magenta",
            "Citric": "yellow",
            "Quinine": "red",
            "Rinse": "lightsteelblue",
            "Lick": "darkgray",
        }

        self.kwargs = kwargs
        self.checks = {}

    def line_signals(self):
        set_pub()

        for col in self.data.columns:
            fig, ax = plt.subplots(1, 1)
            ax.set_xlabel("Time(s)", weight="bold")
            ax.set_ylabel("DF/F", weight="bold")
            ax.patch.set_facecolor = "white"
            Y = self.data.loc[:, col]
            plt.text(0.9, 0.9, f"{col}", transform=ax.transAxes)
            sns.lineplot(ax=ax, x=self.time, y=Y, sort=False)
            plt.show()

    @staticmethod
    def show():
        plt.show()


    def line_fourier(self):
        for col in self.data.columns:
            plt.figure()
            Y = self.data.loc[:, col]
            # X = self.time

            FFT = np.fft.fft(Y)
            new_N = int(len(FFT) / 2)
            f_nat = 1
            new_X = np.linspace(10 ** -12, f_nat / 2, new_N, endpoint=True)
            new_Xph = 1.0 / new_X
            FFT_abs = np.abs(FFT)
            plt.plot(
                new_Xph,
                2 * FFT_abs[0 : int(len(FFT) / 2.0)] / len(new_Xph),
                color="black",
            )
            plt.xlabel("Period ($h$)", fontsize=20)
            plt.ylabel("Amplitude", fontsize=20)
            plt.title("Fourier Transform", fontsize=15)
            plt.grid(True)
            plt.xlim(0, 200)
            plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# _validate.py

Module (nn_utils): 
    Validation functions for SVM. Determine which steps have occured, 
    Prevent steps from occuring before previous attributes have been set
"""

import numpy as np


class _validate:
    @staticmethod
    def _full_predict(pred, true):
        assert (np.unique(pred)) == (np.unique(true))

    @staticmethod
    def _validate_shape(x, y):
        assert x.shape[0] == y.shape[0]

    @staticmethod
    def _check_fitted(model):
        if not hasattr(model, "fit"):
            raise TypeError(f"{model} is not an estimator instance.")

    @staticmethod
    def _num_classes(classes):
        if len(np.unique(classes)) == 2:
            print("Estimator class type: Multivariate")
            return "binary"
        elif len(np.unique(classes)) > 2:
            print("Estimator class type: Multivariate")
            return "multivariate"
        else:
            raise AttributeError(
                "Number of classes do not match binary or multivariate"
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# tracedata.py
Module (data.data_utils): Process traces exported from inscopix trace file.
"""
from __future__ import annotations

from dataclasses import dataclass

import numpy as np
import pandas as pd
import scipy.stats as stats

from canalysis.data.data_utils.file_handler import FileHandler


# %%


@dataclass
class TraceData:
    filehandler: FileHandler = FileHandler
    zscores: pd.DataFrame = None

    def __post_init__(self):
        self.tracedata = self.filehandler.get_tracedata()
        self._clean()
        # Core attributes
        self.signals = self._set_trace_signals()
        self.cells = np.asarray(self.tracedata.columns[1:])
        self.time = np.arange(0, self.tracedata.shape[0] / 10, 0.1)
        self.binsize = self.time[2] - self.time[1]
        self.zscores = self._get_zscores()

    def __repr__(self):
        return type(self).__name__

    @property
    def shape(self):
        return self.signals.shape

    def __hash__(self):
        return hash(repr(self))

    def _get_zscores(self) -> pd.DataFrame:
        zscores = pd.DataFrame(columns=self.signals.columns)
        for cell in self.signals.columns:
            allsignal = self.signals[cell]
            zscore = pd.Series(stats.zscore(allsignal))
            zscores[cell] = zscore
        zscores["time"] = self.time
        return zscores

    def _set_trace_signals(self) -> pd.DataFrame:
        temp = self.tracedata.copy()
        temp.pop("time")
        self.signals = temp
        return temp

    def _clean(self) -> None:

        # When CNMFe is run, we choose cells to "accept", but with manual ROI's every cell is accepted
        # so often that step is skipped. Need some way to check (check_if_accepted)
        _df = self.tracedata.copy()

        def check_if_accepted(_df):
            # If any cells marked as "accepted", use only those cells
            accepted_col = [col for col in _df.columns if " accepted" in col]
            return accepted_col

        accept = check_if_accepted(_df)
        if accept:
            accepted = np.where(_df.loc[0, :] == " accepted")[0]
            _df = _df.iloc[:, np.insert(accepted, 0, 0)]
        _df = _df.drop(0)
        _df = _df.rename(columns={" ": "time"})
        _df = _df.astype(float)
        _df = _df.reset_index(drop=True)
        _df.columns = [column.replace(" ", "") for column in _df.columns]
        _df["time"] = np.round(_df["time"], 2)
        self.tracedata = _df
        return None

    def reorder(self, cols) -> None:
        self.zscores = self.zscores[cols]
        self.zscores['time'] = self.time



from collections.abc import MutableMapping
from canalysis.helpers.wrappers import Singleton


@Singleton
class AllData(MutableMapping):
    """
    Custom mapping that works with properties from mutable object. Used to store each
    instance
    of data to iterate over and provide additional functionality.

    ..: Usage:
    alldata = AllData(function='xyz')
    and
    d.function returns 'xyz'
    """

    def __init__(self, *args, **kwargs):
        self.__dict__.update(*args, **kwargs)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __str__(self):
        return "\n".join(
            f"{key} - {len(value)} sessions." for key, value in self.__dict__.items()
        )

    def __repr__(self):
        return "\n".join(
            f"{key} - {len(value)} sessions." for key, value in self.__dict__.items()
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
#process_data.py

Module(analysis): Stats class for PCA and general statistics/output.
"""
from __future__ import annotations

from typing import Iterable, Any, ClassVar
import pandas as pd

from utils import funcs


def sort_by_value(data):
    pass


class ProcessData:
    def __init__(self, data,):
        self.data: ClassVar = data
        self.signals: pd.DataFrame = data.tracedata.signals
        self.zscores: pd.DataFrame = data.tracedata.zscores
        self.time: pd.Series | Iterable[Any] = data.tracedata.time
        self.cells: Iterable[Any] = data.tracedata.cells
        self.avgs: dict = data.nr_avgs
        self.session: str = data.filehandler.session
        self.trial_times: dict = data.eventdata.trial_times
        self.timestamps: dict = data.eventdata.timestamps
        self.antibouts = self.get_antibouts()
        self.sponts = self.get_sponts()

    def get_antibouts(self) -> pd.DataFrame | pd.Series:
        antibouts = pd.DataFrame()
        for interv in funcs.interval(self.timestamps["Lick"], gap=10, outer=True):
            df = self.signals.loc[(self.time > (interv[0])) & (self.time < (interv[1]))]
            antibouts = pd.concat([antibouts, df], axis=0)
        return antibouts

    def get_sponts(self) -> pd.DataFrame | pd.Series:
        sponts = pd.DataFrame()
        for interv in funcs.interval(self.timestamps["Lick"], gap=30, outer=True):
            df = self.signals[(self.time > (interv[0])) & (self.time < (interv[1]))]
            sponts = pd.concat([sponts, df], axis=0)
        return sponts



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
#helpers.py

Module (graph): General functions for graphing.
"""
from __future__ import annotations
from typing import Tuple, Optional

import matplotlib
from matplotlib import rcParams, lines
import matplotlib.pyplot as plt

import numpy as np
from sklearn.model_selection import learning_curve


def update_rcparams():
    # Function to set some easy params and avoid some annoying bugs
    rcParams.update(
            {
                "font.weight": "bold",
                "font.family": "Arial",
                'font.sans-serif': "Arial",
                "axes.labelweight": "bold",
                "xtick.major.width": "1.3",
                "axes.facecolor": "w",
                "axes.labelsize": 10,
                "lines.linewidth": 1,
                'animation.ffmpeg_path': r'/c/ffmpeg/bin/ffmpeg',
                'scatter.edgecolors': None
            }

    )


def get_handles_from_dict(
    color_dict: dict,
    markersize: Optional[int] = 5,
    marker: Optional[str] = 'o',
    **kwargs
) -> Tuple[list, list]:
    """
    Get matplotlib handles for input dictionary.
    Args:
        markersize ():
        color_dict (dict): Dictionary of event:color k/v pairs.
        marker (str): Shape of scatter point, default is circle.
    Returns:
        proxy (list): matplotlib.lines.line2D appended list.
        label (list): legend labels for each proxy.
    """
    proxy, label = [], []
    for t, c in color_dict.items():
        proxy.append(
                lines.Line2D(
                        [0],
                        [0],
                        marker=marker,
                        markersize=markersize,
                        markerfacecolor=c,
                        markeredgecolor="None",
                        linestyle="None",
                        **kwargs
                ))
        label.append(t)
    return proxy, label


def confidence_ellipse(x, y, ax, n_std=1.8, facecolor="none", **kwargs) -> matplotlib.figure.Axes:
    """
    Create a covariance confidence ellipse of `x` and `y`
            
    Parameters
    ----------
    x, y : array_like, shape (n, )
        Input data.
    ax : matplotlib.axes.Axes
        The axes object to draw the ellipse into.
    n_std : float
        The number of standard deviations to determine the ellipse's radiuses.
    facecolor : str
        Color of background.
    Returns
    -------
    matplotlib.patches.Ellipse
    Other parameters
    ----------------
    kwargs : `~matplotlib.patches.Patch` properties
    """
    from matplotlib.patches import Ellipse
    import matplotlib.transforms as transforms

    if x.size != y.size:
        raise ValueError("x and y must be the same size")
    cov = np.cov(x, y)
    pearson = cov[0, 1] / np.sqrt(cov[0, 0] * cov[1, 1])
    # obtain the eigenvalues (special case)
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    ellipse = Ellipse(
            (0, 0),
            width=ell_radius_x * 2,
            height=ell_radius_y * 2,
            facecolor=facecolor,
            **kwargs
    )
    # standard deviation of x
    scale_x = np.sqrt(cov[0, 0]) * n_std
    mean_x = np.mean(x)
    # standard deviation of y
    scale_y = np.sqrt(cov[1, 1]) * n_std
    mean_y = np.mean(y)
    transf = (
        transforms.Affine2D()
            .rotate_deg(45)
            .scale(scale_x, scale_y)
            .translate(mean_x, mean_y)
    )
    ellipse.set_transform(transf + ax.transData)
    return ax.add_patch(ellipse)


def plot_learning_curve(
    estimator,
    title,
    X,
    y,
    axes=None,
    ylim=None,
    cv=None,
    n_jobs=None,
    train_sizes=np.linspace(0.1, 1.0, 5),
):
    """
    Generate 3 plots: the test and training learning curve, the training
    samples vs fit times curve, the fit times vs score curve.

    Parameters
    ----------
    estimator : estimator instance
        An estimator instance implementing `fit` and `predict` methods which
        will be cloned for each validation.

    title : str
        Title for the chart.

    X : array-like of shape (n_samples, n_features)
        Training vector, where ``n_samples`` is the number of samples and
        ``n_features`` is the number of features.

    y : array-like of shape (n_samples) or (n_samples, n_features)
        Target relative to ``X`` for classification or regression;
        None for unsupervised learning.

    axes : array-like of shape (3,), default=None
        Axes to use for plotting the curves.

    ylim : tuple of shape (2,), default=None
        Defines minimum and maximum y-values plotted, e.g. (ymin, ymax).

    cv : int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

          - None, to use the default 5-fold cross-validation,
          - integer, to specify the number of folds.
          - :term:`CV splitter`,
          - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`StratifiedKFold` used. If the estimator is not a classifier
        or if ``y`` is neither binary nor multiclass, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validators that can be used here.
    n_jobs : int or None, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.
    train_sizes : array-like of shape (n_ticks,)
        Relative or absolute numbers of training examples that will be used to
        generate the learning curve. 
    """
    if axes is None:
        _, axes = plt.subplots(1, 3, figsize=(20, 5))

    axes[0].set_title(title)
    if ylim is not None:
        axes[0].set_ylim(*ylim)
    axes[0].set_xlabel("Training examples")
    axes[0].set_ylabel("Score")

    train_sizes, train_scores, test_scores, fit_times, _ = learning_curve(
            estimator,
            X,
            y,
            cv=cv,
            n_jobs=n_jobs,
            train_sizes=train_sizes,
            return_times=True,
    )
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    fit_times_mean = np.mean(fit_times, axis=1)
    fit_times_std = np.std(fit_times, axis=1)

    # Plot learning curve
    axes[0].grid()
    axes[0].fill_between(
            train_sizes,
            train_scores_mean - train_scores_std,
            train_scores_mean + train_scores_std,
            alpha=0.1,
            color="r",
    )
    axes[0].fill_between(
            train_sizes,
            test_scores_mean - test_scores_std,
            test_scores_mean + test_scores_std,
            alpha=0.1,
            color="g",
    )
    axes[0].plot(
            train_sizes, train_scores_mean, "o-", color="r", label="Training score"
    )
    axes[0].plot(
            train_sizes, test_scores_mean, "o-", color="g", label="Cross-validation score"
    )
    axes[0].legend(loc="best")

    # Plot n_samples vs fit_times
    axes[1].grid()
    axes[1].plot(train_sizes, fit_times_mean, "o-")
    axes[1].fill_between(
            train_sizes,
            fit_times_mean - fit_times_std,
            fit_times_mean + fit_times_std,
            alpha=0.1,
    )
    axes[1].set_xlabel("Training examples")
    axes[1].set_ylabel("fit_times")
    axes[1].set_title("Scalability of the model")

    # Plot fit_time vs score
    fit_time_argsort = fit_times_mean.argsort()
    fit_time_sorted = fit_times_mean[fit_time_argsort]
    test_scores_mean_sorted = test_scores_mean[fit_time_argsort]
    test_scores_std_sorted = test_scores_std[fit_time_argsort]
    axes[2].grid()
    axes[2].plot(fit_time_sorted, test_scores_mean_sorted, "o-")
    axes[2].fill_between(
            fit_time_sorted,
            test_scores_mean_sorted - test_scores_std_sorted,
            test_scores_mean_sorted + test_scores_std_sorted,
            alpha=0.1,
    )
    axes[2].set_xlabel("fit_times")
    axes[2].set_ylabel("Score")
    axes[2].set_title("Performance of the model")

    return plt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
#analysis_funcs.py

Module(analysis_utils): Functions to assist in data analysis
"""
from __future__ import annotations

import logging
from typing import Optional, Any, Iterable
import numpy as np
import pandas as pd
from scipy.stats import stats
from utils import funcs


def map_colors(
    iterable: Iterable[Any],
    colors_dict: dict = None
) -> list:
    """
    Given dataframe and dictionary, maps value of dictionary to event of dataframe column.
    Parameters
    """
    if not colors_dict:
        colors_dict = {
            'grooming': 'green',
            'entry': 'blue',
            'eating': 'red'
        }
    return [colors_dict[x] for x in iterable]



def get_stats(self) -> pd.DataFrame | None:
    stats_df = pd.DataFrame(
            columns=[
                "File",
                "Cell",
                "Stimulus",
                "Trial",
                "Baseline (mean)",
                "Baseline (st_dev)",
                "Signal Timestamps (start, stop)",
                "Baseline Timestamps (start, stop)",
                "Shifted?",
                "deltaF/F",
                "Significant?",
            ]
    )
    summary = pd.DataFrame(columns=["File", "Stimulus", "Num Trials"])
    raw_df = pd.DataFrame(columns=self.cells)
    count = 0
    stats_list = []
    for cell in self.signals.columns:
        count += 1
        cell_list = []
        for stim, times in self.trial_times.items():
            ntrials = len(times)
            trial_list = []
            for iteration, trial in enumerate(times):
                # Index to analysis window
                data_ind = np.where((self.time > trial) & (self.time < trial + 5))[
                    0
                ]
                # Index to baseline window (current: 4s pre stim)
                bltime_ind = np.where(
                        (self.time > trial - 4) & (self.time < trial)
                )[0]
                bltime = self.time[bltime_ind]
                bltime = [min(bltime), max(bltime)]
                # Baseline statistics
                bl_signal = np.array(self.signals[bltime_ind, count])
                mean_bl = np.mean(bl_signal)
                stdev_bl = np.std(bl_signal)
                c_interv = mean_bl + stdev_bl * 2.58
                # Get calcium trace & time
                signal = np.array(self.signals[data_ind, count])
                # Get peak signal & time
                peak_signal = max(signal)
                peak_ts = self.signals[
                    self.signals[cell] == peak_signal, "time"
                ].iloc[0]
                if peak_ts <= trial + 2.4:
                    peak_ts = max(bltime) + 2.4
                    shift = "yes"
                else:
                    shift = "no"
                # Get window 1s centered at peak
                peak_window_ind = funcs.get_peak_window(self.time, peak_ts)
                time_lower = self.time[peak_window_ind[0]]
                time_upper = self.time[peak_window_ind[1]]
                response_window = np.array(
                        self.signals[peak_window_ind[0]: peak_window_ind[1], cell]
                )
                window_ts = [time_lower, time_upper]
                mean_mag = np.mean(response_window)
                dff = (mean_mag - mean_bl) / mean_bl
                # if DF/F is significant, output concat new dataframe
                # with raw values from this trial
                if dff >= c_interv:
                    d = [stim, "Trial {}".format(iteration + 1)]
                    d.extend(["-" for _ in range(len(self.signals[1:]))])
                    track = len(raw_df)
                    sig = "Significant"
                    raw_df.loc[track] = d
                    baseline_df = self.signals[
                        self.time.between(min(bltime), time_upper)
                    ]
                    raw_df = pd.concat([raw_df, baseline_df])
                else:
                    sig = "ns"
                # Pack analysis into list
                iteration_list = [
                    self.session,
                    cell,
                    stim,
                    f"Trial {iteration + 1}",
                    mean_bl,
                    stdev_bl,
                    window_ts,
                    bltime,
                    shift,
                    dff,
                    sig,
                ]
                trial_list.append(iteration_list)
            if cell == self.signals[0]:
                summary_list = [self.session, stim, ntrials]
                track = len(summary)
                summary.loc[track] = summary_list
            cell_list.append(trial_list)
        stats_list.append(cell_list)
    flaten_stats = funcs.flatten(funcs.flatten(stats_list))
    # Make dataframe from flattened list of analysis
    for ind, d in enumerate(flaten_stats):
        track = len(stats_df)
        stats_df.loc[track] = d
    logging.info("Stats successfully completed.")

    if self.outpath:
        print("Outputting data to Excel...")
        with pd.ExcelWriter(self.outpath / "/" / "_statistics.xlsx") as writer:
            summary.to_excel(writer, sheet_name="Summary", index=False)
            raw_df.to_excel(writer, sheet_name="Raw Data", index=False)
            # lickstats.to_excel(
            #     writer, sheet_name='Lick cells', index=False)
            stats_df.to_excel(writer, sheet_name="Trial Stats", index=False)
        print(" statistics successfully transferred to Excel!")
        return None
    else:
        return stats_df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
#data.py

Module: Classes for data processing.
"""

from __future__ import annotations
import logging
from dataclasses import dataclass, field, InitVar
from typing import ClassVar, Optional
import pandas as pd
from canalysis.data.containers.all_data import AllData
from canalysis.data.containers.trace_data import TraceData
from canalysis.data.containers.taste_data import TasteData
from canalysis.data.containers.event_data import EventData
from canalysis.data.containers.eating_data import EatingData
from canalysis.data.data_utils.file_handler import FileHandler
from canalysis.graphs.graph_utils import Mixins
from canalysis.helpers import excepts as e
from canalysis.helpers import funcs

logger = logging.getLogger(__name__)


@dataclass
class CalciumData(Mixins.CalPlots):
    """
    General holder class for all trace/event related data, with additional
    storage for each session.
    """

    __filehandler: FileHandler
    color_dict: dict
    adjust: Optional[int | float | None] = 0
    doevents: Optional[bool] = True
    doeating: Optional[bool] = True
    tracedata: TraceData = field(init=False)
    eventdata: EventData = field(init=False)
    tastedata: TasteData = field(init=False)
    alldata: ClassVar[AllData] = AllData.Instance()

    def __post_init__(self):
        # Instance info
        self.date = self.__filehandler.date
        self.animal = self.__filehandler.animal
        self.data_dir = self.__filehandler.directory
        self.session = self.__filehandler.session
        self.doevents: Optional[bool] = self.doevents
        self.doeating: Optional[bool] = self.doeating
        # Core data
        self.tracedata: TraceData = TraceData(self.__filehandler)
        if self.doevents is True:
            self.eventdata: EventData = EventData(self.__filehandler, self.color_dict, self.tracedata.time)
            self.nr_avgs = self._get_nonreinforced_means()
            self.tastedata: TasteData = TasteData(
                    self.tracedata.zscores,
                    self.tracedata.time,
                    self.eventdata.timestamps,
                    self.eventdata.trial_times,
                    self.nr_avgs,
                    self.color_dict
            )
        else:
            logging.info("skipping events")
        if self.doeating is True:
            if self.__filehandler.eatingname is not None:
                self.eatingdata: EatingData = EatingData(
                    self.__filehandler,
                    self.tracedata,
                    self.color_dict,
                    self.adjust
            )
        self.cells = self.tracedata.cells
        self._authenticate(doevents=self.doevents)
        self._add_instance()

    @classmethod
    def __len__(cls):
        return len(cls.alldata.keys())

    def __repr__(self):
        return type(self).__name__

    @staticmethod
    def keys_exist(element, *keys):
        return funcs.keys_exist(element, *keys)

    @property
    def size(self):
        return len(self.tracedata.cells)

    def reset_tastedata(self):
        self.tastedata = TasteData(
                    self.tracedata.zscores,
                    self.tracedata.time,
                    self.eventdata.timestamps,
                    self.eventdata.trial_times,
                    self.nr_avgs,
                    self.color_dict
                )

    def _authenticate(self, doevents: bool | None = True):
        if not isinstance(self.tracedata, TraceData):
            raise e.DataFrameError("Trace data must be a dataframe")
        if doevents is True:
            if not isinstance(self.eventdata, EventData):
                raise e.DataFrameError("Event data must be a dataframe.")
        if not any(
                x in self.tracedata.signals.columns for x in
                ["C0", "C00", "C000", "C0000"]
        ):
            logging.debug(f"{self.tracedata.signals.head()}")
            raise AttributeError(
                f"No cells found in DataFrame: " f"{self.tracedata.signals.head()}"
            )
        return None

    def _add_instance(self):
        my_dict = type(self).alldata
        if self.keys_exist(my_dict, self.animal, self.date):
            logging.info(f"{self.animal}-{self.date} already exist.")
        elif self.keys_exist(my_dict, self.animal) and not self.keys_exist(
                my_dict, self.date
        ):
            my_dict[self.animal][self.date] = self
            logging.info(f"{self.animal} exists, {self.date} added.")
        elif not self.keys_exist(my_dict, self.animal):
            my_dict[self.animal] = {self.date: self}
            logging.info(f"{self.animal} and {self.date} added")
        return None

    def _get_nonreinforced_means(
            self,
    ) -> dict:
        """
        Get a dictionary (Cell: mean) of means for non-reinforced lick events.
        Returns
        -------
        Cell: mean dictionary.
        """
        nr_signal = self.tracedata.zscores.loc[
            self.tracedata.zscores["time"].isin(self.eventdata.nonreinforced)
        ].drop("time", axis=1)
        avgs = {}
        for column in nr_signal.columns:
            mean = (nr_signal[column]).mean()
            avgs[column] = mean
        return avgs

    def get_signal_bycell(
            self,
            i
    ) -> list[pd.Series]:
        """return a list of signal values via cell integer indexing (0 through N cells"""
        return list(self.tracedata.signals.iloc[:, i])

    def combine(self, eatingevent: list = None, stimevent: list = None) -> tuple[pd.DataFrame, pd.Series]:
        eating, eatingcolors = self.eatingdata.get_signals_from_events(eatingevent)
        stim, stimcolors = self.tastedata.get_signals_from_events(stimevent)
        return pd.concat([eating, stim], axis=0), pd.concat([eatingcolors, stimcolors], axis=0)
